# Now what you only need to do for construct a indicator is to call
# BaseIndicator(RSI(24*60),['ask_close'])
# for a simplified calling: BI(RSI,24*60,['ask_open','ask_close'])
import itertools

from indicator.wrappers import BaseIndicator
from ..indicators import classes


def BI(indicator_class, params, args):
    """
                example:- BI(RSI,24*60,['ask_open','ask_close'])
                        - BI(MACD,[24*60,24*60*5],['ask_open'])
    :param indicator_class:  The class of Indicator
    :param params: params or param use to create instance
    :param args: args or arg passing to `to_next` , you can use either `list` or `dict`
    :return: BaseIndicator,with name in Indicator(param1,param2)[arg1,arg2....] like `RSI(60)[ask_open,ask_close]`
    """
    if type(params) is tuple:
        _params = list(params)
    elif type(params) is not list:
        _params = [params]
    else:
        _params = params

    # 无论如何都换成 list
    if type(args) is tuple:
        _args = list(args)
    elif type(args) is not list:
        _args = [args]
    else:
        _args = args
    instance = indicator_class(*_params)

    param_str = f"({str(_params)[1:-1]})"
    arg_str = f"[{','.join(_args)}]"

    return BaseIndicator(instance, _args, f"{indicator_class.__name__}{param_str}{arg_str}")


def BIG(indicator_class, params_list, args_list):
    indicators = []
    pa_iter = itertools.product(params_list, args_list)
    for params, args in pa_iter:
        indicators.append(BI(indicator_class, params, args))

    return indicators


def BIN(indicator_class_name, params, args):
    assert indicator_class_name in classes.keys(), f'indicator:{indicator_class_name} not exists.'
    return BI(classes[indicator_class_name], params, args)


def BIS(indicator_string: str):
    # TODO: Check the Format
    [name, r] = indicator_string.split('(')
    [params, remain] = r.split(')')
    params = eval(params)
    args = remain.replace('[', '').replace(']', '').split(',')
    return BIN(name, params, args)


def parseTechniqueStrict(technique: str):
    """
        for example : 0.125 < RSI(30)[ask_open,ask_close] < 0.45
    :param technique:
    :return:
    """
    format = technique.split('<')
    indicator = BIS(format[1])
    low = float(format[0])
    high = float(format[2])
    return {
        "low": low,
        "high": high,
        "indicator": indicator
    }



from indicator.indicators import MACD, SimpleSlope


class MACDSlope:
    def __init__(self, slow_period, fast_period, slope_period):
        self.macd = MACD.MACD(slow_period, fast_period)
        self.slope = SimpleSlope.SimpleSlope(slope_period)
        self.line = []

    def to_next(self, slow_data, fast_data):
        macd = self.macd.to_next(slow_data, fast_data)
        if macd is None:
            self.line.append(None)
            return None
        slope = self.slope.to_next(macd)
        if slope is None:
            self.line.append(None)
            return None
        self.line.append(slope)
        return slope



export = MACDSlope


from tqdm import tqdm
import os
import requests

frequencies = ['5min', '15min', '30min', '1h', '4h', '6h', '1D', '5D', '15D', '1W', '1M', '1Y']
base_url = 'https://ewr1.vultrobjects.com/fxcm/'

bundle_name = 'ratio'

affs = ['_' + f + '.h5' for f in frequencies]
affs.append('.h5')

print('start downloading bundle files.')

for aff in affs:
    filename = bundle_name + aff
    url = base_url + filename
    # Streaming, so we can iterate over the response.

    print(f'downloading {filename}')

    if os.path.exists(os.path.join('data', 'bundle', filename)):
        print(f'{os.path.join("data", "bundle", filename)} has already exists.'
              f'If you want to download it again,please delete it manually.')
        continue

    r = requests.get(url, stream=True)
    # Total size in bytes.
    total_size = int(r.headers.get('content-length', 0))
    block_size = 1024  # 1 Kibibyte
    t = tqdm(total=total_size, unit='iB', unit_scale=True)
    with open(os.path.join('data', 'bundle', filename), 'wb') as f:
        for data in r.iter_content(block_size):
            t.update(len(data))
            f.write(data)
    t.close()
    if total_size != 0 and t.n != total_size:
        print(f"ERROR, something went wrong when downloading: {filename}")



cimport numpy as np
cdef class Bar:
     cdef readonly np.float64_t end
     cdef readonly np.float64_t start
     cdef readonly np.float64_t ask_open
     cdef readonly np.float64_t ask_close
     cdef readonly np.float64_t ask_low
     cdef readonly np.float64_t ask_high
     cdef readonly np.float64_t bid_open
     cdef readonly np.float64_t bid_close
     cdef readonly np.float64_t bid_low
     cdef readonly np.float64_t bid_high
     cdef readonly np.float64_t volume



# Entry Point of our new trader
# 这个模块的主要目的是获取命令行参数并且调用相应的工具进行回测
import datetime
import os
import traceback

import click
import pandas as pd
import prettytable as pt

import util
import quotation
import runner
from indicator import parseTechniqueStrict

import sys


# We append strategy dir.
sys.path.append('./strategies')


@click.command()
@click.option('-b', '--bundle', default='ratio', help="History Data Bundle")
@click.option('-s', '--strategy', default='range', help="strategy Module")
@click.option('-S', '--start_datetime', default='2012-01-01', help="start date")
@click.option('-E', '--end_datetime', default='2019-01-01', help="start date")
@click.option('-I', '--initial_cash', default=10000, help="initial cash", type=float)
@click.option('-C', '--commission_rate', default=0, help="commission rate", type=float)
@click.option('-O', '--output_file', default=None, help=" select an output file to save the result data in HDF format."
                                                        "In default,we will save as "
                                                        "data/results/backtest-YYYY-MM-DD-HH-mm-ss.h5"
              )
@click.option(
    '-D',
    '--defines',
    multiple=True,
    help="Define a name to be bound in the namespace before executing"
         " the algotext. For example '-D name=value'. The value may be any python"
         " expression. These are evaluated in order so they may refer to previously"
         " defined names.",
)
@click.option('-t', '--technique', default=None, type=str,
              help=" Use Technique Strict, format is low < Indicator_name < High",
              multiple=True)
@click.option('-f', '--frequency', default=None, type=str)
@click.option('-i', '--instrument', default=None, help='instrument,default is to be set by the strategy')
def main(bundle, strategy, start_datetime, end_datetime, initial_cash, commission_rate, output_file, defines,
         instrument, frequency, technique):
    str_module = __import__(strategy)

    param_dict = {}

    for define in defines:
        key, value = define.split('=')
        value = eval(value)
        param_dict[key] = value

    """[summary]
        加载行情数据
        注意： 行情数据从本地配置
    """
    qt = quotation.LocalQuotation()

    ctx = runner.Context(quotation=qt, initial_cash=initial_cash, commission_rate=commission_rate,
                         start_date=start_datetime,
                         end_date=end_datetime)
    # intialize the context
    str_module.init(ctx)

    if instrument is not None:
        if util.forex.is_supported(instrument):
            ctx.instrument = instrument
        else:
            print(' warning: Invalid instrument specified.Use the default setting by the strategy')

    if frequency is not None:
        if frequency in quotation.frequency.supported:
            ctx.frequency = frequency
        elif ctx.frequency is not None:
            print('waring: Invalid frequency specified.Use the default setting by the strategy')
        else:
            ctx.frequency = '1D'
            print('warning: Invalid frequency specified.Use the default of 1D')

    technique_list = []

    if technique is not None:
        for t in technique:
            ti = parseTechniqueStrict(t)
            technique_list.append(ti)
            ctx.parameters_instance[ti["indicator"].name] = f'({ti["low"]},{ti["high"]})'

    for key, param in ctx.parameters.items():
        if key in param_dict:
            assert param.check_valid(param_dict[key]), f'Should low= {param.low} <=value= ' \
                f'{param_dict[key]} < high= {param.high}'
            ctx.parameters_instance[key] = param_dict[key]
        else:
            ctx.parameters_instance[key] = param.default_value

    print('Parameters:')

    tb = pt.PrettyTable(['param', 'value'])
    tb.align["param"] = 'l'
    tb.align["value"] = 'r'
    tb.padding_width = 5
    tb.hrules = pt.ALL
    ctx.parameters_instance['str_name'] = ctx.str_name
    ctx.parameters_instance['start_datetime'] = ctx.start_date
    ctx.parameters_instance['end_datetime'] = ctx.end_date
    ctx.parameters_instance['instrument'] = ctx.instrument
    ctx.parameters_instance['period'] = ctx.frequency
    ctx.parameters_instance['commission_rate'] = ctx.commission_rate
    ctx.parameters_instance['initial_cash'] = ctx.initial_cash
    for key, param in ctx.parameters_instance.items():
        tb.add_row([key, param])

    print(tb)

    print('Loading Quotation...')

    qt.load(ctx.instrument, ctx.frequency, start_datetime, end_datetime, bundle)

    bt = runner.BackTestRunner(str_module, ctx)

    print('Start to run backtest.')

    bt.attach_techniques(technique_list)
    bt.run()
    r = bt.conclude()

    print('Backtest finished.')

    if not r:
        print('No Trade Was Made,So No Result Will be Output.')
        return False

    units = {
        "win_rate": "%",
        "net_ret": "%",
        "max_drawdown": "%"
    }

    round_i = {
        "win_rate": 2,
        "net_ret": 2,
        "max_drawdown": 2,
        "sharp_ratio": 4,
        "plr": 4,
    }

    tb = pt.PrettyTable(['Cost Function', 'Value'])
    tb.align["Cost Function"] = 'l'
    tb.align["Value"] = 'r'
    tb.padding_width = 5
    tb.hrules = pt.ALL
    print('backtest result:')
    for key, value in bt.cost_functions.items():

        u = ''

        if key in units:
            u = units[key]
        v = value
        if key in round_i:
            v = round(v, round_i[key])
        tb.add_row([key, f'{v}{u}'])

    print(tb)

    if output_file is None:
        output_file = os.path.join('data', 'results',
                                   'backtest-%s.h5' % (datetime.datetime.now().strftime('%Y-%m-%d_%H:%m:%S')))

    try:
        with pd.HDFStore(output_file) as hdf:
            hdf.put('trades', bt.trade_df)
            hdf.put('pnl', bt.pnl)
            hdf.put('params', pd.DataFrame([ctx.parameters_instance]))
            hdf.put('cost_functions', pd.DataFrame([bt.cost_functions]))
            print(f'result has been saved to {output_file}')
    except Exception as e:
        traceback.print_exc()


if __name__ == '__main__':
    main()



from .MACD import MACD
from .range import Range


class NMACD:
    def __init__(self, fast_p, slow_p):
        self.macd = MACD(fast_p, slow_p)
        self.range = Range(slow_p)
        self.line = []

    def to_next(self, price, high, low):
        macd = self.macd.to_next(price, price)
        r = self.range.to_next(high, low)
        if macd is not None and r is not None:
            self.line.append(macd / r)
            return self.line[-1]
        else:
            self.line.append(None)
            return None


export = NMACD



import click
import os
import datetime
import quotation
import runner

import pandas as pd
import traceback
import itertools

import util.forex

import sys

def save_results(param_dict, fixed_param_dict, ctx, results, filename):
    res_df = pd.DataFrame(results)
    res_df = res_df.sort_values(by="sharp_ratio", ascending=False)
    print(res_df)

    meta_info = pd.DataFrame([{
        "str_name": ctx.str_name,
        "frequency": ctx.frequency,
        "instrument": ctx.instrument,
        "initial_cash": ctx.initial_cash,
        "commission_rate": ctx.commission_rate,
        "start_datetime": ctx.start_date,
        "end_datetime": ctx.end_date
    }])

    mutable_params = list(param_dict.keys())
    fixed_params = list(fixed_param_dict.keys())

    for param in mutable_params:
        if param in fixed_params:
            mutable_params.remove(param)

    mutable_params = pd.DataFrame({
        'names': mutable_params,
    })

    all_params = pd.DataFrame({
        'names': list(ctx.parameters.keys())
    })

    if filename is None:
        filename = os.path.join('data', 'results',
                                'scan-%s.h5' % (datetime.datetime.now().strftime('%Y-%m-%d_%H:%m:%S')))
    try:
        res_df.to_hdf(filename, 'result')
        meta_info.to_hdf(filename, 'meta')
        mutable_params.to_hdf(filename, 'mutable_params')
        all_params.to_hdf(filename, 'all_params')
        print(f'result has been saved to {filename}')
    except Exception:
        traceback.print_exc()


def process_params(ranges, defines):
    param_dict = {}
    fixed_param_dict = {}

    for define in ranges:
        key, value = define.split('=')
        value = int(value)
        param_dict[key] = value

    for define in defines:
        key, value = define.split('=')
        value = eval(value)
        fixed_param_dict[key] = value

    return param_dict, fixed_param_dict


def generate_params(ctx, param_dict, fixed_param_dict):
    instances = dict()
    instances['initial_cash'] = [ctx.initial_cash]
    instances['instrument'] = [ctx.instrument]
    instances['commission_rate'] = [ctx.commission_rate]

    # 在instance里面生成一组参数
    for key, param in ctx.parameters.items():
        if key in fixed_param_dict:
            # 首选项： 如果有固定的参数，我们一定把它设置为固定的。
            instances[key] = [fixed_param_dict[key]]
        elif key in param_dict:
            instances[key] = param.generate(param_dict[key])
        else:
            instances[key] = [param.default_value]
    return instances


def product_params(instances):
    # 考虑把他们都乘起来

    for key, values in instances.items():
        instances[key] = [(key, v) for v in values]

    #
    # 这里的结构是： [((k1,v1_1),(k2,v2_1),...),((k1,v1_1),(k2,v2_2),...)]
    prod = list(itertools.product(*instances.values()))

    dict_list = []

    for t in prod:
        d = {}
        for e in t:
            k, v = e
            d[k] = v
        dict_list.append(d)

    return dict_list


sys.path.append('./strategies')


@click.command()
@click.option('-b', '--bundle', default='ratio', help="History Data Bundle")
@click.option('-s', '--strategy', default='range', help="strategy Module")
@click.option('-S', '--start_datetime', default='2014-01-01', help="start date")
@click.option('-E', '--end_datetime', default='2018-01-01', help="start date")
@click.option('-I', '--initial_cash', default=10000, help="initial cash", type=float)
@click.option('-C', '--commission_rate', default=0, help="commission rate", type=float)
@click.option('-P', '--process_count', default=None, help="process count, default is cpu count", type=int)
@click.option('-O', '--output_file', default=None, help=" select an output file to save the result data in HDF format."
                                                        "In default,we will save as "
                                                        "data/results/scan-YYYY-MM-DD-HH-mm-ss.h5")
@click.option('-i', '--instrument', default=None, help='instrument,default is to be set by the strategy')
@click.option(
    '-D',
    '--defines',
    multiple=True,
    help="Set fixed values for scanning of the certain parameter before executing"
         " computing . For example '-Dname=value'. The value will be used to fix the parameter you choosed, and the "
         "selected parameters will be removed for the mutable parameters"
)
@click.option(
    '-R',
    '--ranges',
    multiple=True,
    help="Set range for scanning of the certain parameter before executing"
         " computing . For example '-R name=value'. the value should be integers "
         "which times you want to calculate repeatedly for "
         " the certain parameter"
)
@click.option('-f', '--frequency', default=None, type=str)
def main(bundle, strategy, start_datetime, end_datetime, initial_cash, commission_rate, ranges, process_count,
         output_file, defines, instrument, frequency):
    str_module = __import__(strategy)

    param_dict, fixed_param_dict = process_params(ranges, defines)

    """[summary]
        加载行情数据
        注意： 行情数据从本地配置
    """
    qt = quotation.LocalQuotation()

    ctx = runner.Context(quotation=qt, initial_cash=initial_cash, commission_rate=commission_rate,
                         start_date=start_datetime, end_date=end_datetime)

    # intialize the context
    str_module.init(ctx)

    if instrument is not None:
        if util.forex.is_supported(instrument):
            ctx.instrument = instrument
        else:
            print(' warning: Invalid instrument specified.Use the default setting by the strategy')

    if frequency is not None:
        if frequency in quotation.frequency.supported:
            ctx.frequency = frequency
        elif ctx.frequency is not None:
            print('waring: Invalid frequency specified.Use the default setting by the strategy')
        else:
            ctx.frequency = '1D'
            print('warning: Invalid frequency specified.Use the default of 1D')

    qt.load(ctx.instrument, ctx.frequency, start_datetime, end_datetime, bundle)

    instances = generate_params(ctx, param_dict, fixed_param_dict)

    dict_list = product_params(instances)

    print('start to scan')

    try:
        results = runner.run_scan(strategy, ctx, dict_list, process_count)

        save_results(param_dict, fixed_param_dict, ctx, results, output_file)

    except Exception:
        traceback.print_exc()
    finally:
        qt.close_share()


if __name__ == '__main__':
    main()



class Quotation:
    def __init__(self,instrument,frequency):
        self.instrument = instrument
        self.frequency = frequency
    def load(self):
        raise NotImplementedError("Load Quotation is not implemented")
    def next(self,timeout=-1):
        """[summary]
            Should Return the next one quotation in synchron.
            When timeout is hit,should raise the TimeoutException.
        Raises:
            NotImplementedError: [description] This method should not be called directly.
        """
        raise NotImplementedError("Next Quotation is not implemented")
    def async_next(self,callback):
        """[summary]
            Should Return the next one quotation in async mode.
            The Argument is callack function.
        Arguments:
            callback {function} -- [description]
        
        Raises:
            NotImplementedError: [description]
        """
        raise NotImplementedError("Async Next Quotation is not implemented")